#!/bin/bash
#
# mkbuild: a build system for iPhone OS projects
# version 1.0b3 (February 26, 2009)
#
# Copyright (c) 2009 Benjamin Ragheb <ben@benzado.com>
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

# GENERAL OVERVIEW

# The purpose of this script is to provide a simple build numbering system for
# iPhone OS projects (though it could probably be adapted to any project using
# Xcode).  By executing a single command, you can produce a clean build of
# your code and also tag the source so that it may be recreated if necessary.

# To create a build, it tags your repository (both subversion and git are
# supported), then checks out that tag to a RAM disk.  The build is performed
# on the RAM disk, and then needed build products are zipped into an output
# directory.  Exactly what is built is determined by the do_build function
# defined in a build-config.sh file in the current directory.

# For example, I run the script from a `Builds` directory within my project.
# It creates a tag, say `build-0006` (because it determined from existing tags
# that is the next build number).  It performs the build and then writes some
# files to a directoy also called `build-0006` inside my `Builds` directory.
# For iPhone projects, I create both AdHoc and AppStore-ready zip files, so I
# can test one and upload the other.

# The script sets the Xcode configuration property BUILD_NUMBER, which you can
# use inside your project to display the exact build number.


# CONFIGURATION DEFAULTS
# Don't edit here; override these in your build-config.sh file.

# SCM must be either 'git' or 'svn'
SCM=none

# The source control system's tag name is derived by prepending the value of
# TAG_PREFIX to the build number. The build number is padded with zeroes.
TAG_PREFIX=build-

# To perform a build, mkbuild creates a RAM disk.  This is the default RAM
# disk size, in megabytes.
RAMDISK_SIZE_MB=10

# Final artifacts of the build will be written to a directory with the same
# name as the build tag, as a child of the OUTPUT_ROOT directory.
OUTPUT_ROOT=$PWD

# Path to the trunk and tags directories in your subversion repository.
SVN_PATH_TRUNK=trunk
SVN_PATH_TAGS=tags


# BUILD FUNCTIONS
# Call these from the do_build function in your build-config.sh file.


# Performs a build on the RAM disk
make_app() {
	local XCODE_TARGET=$1
	local XCODE_CONFIGURATION=$2
	
	local XCODE_OUTPUT=$OUTPUT_PATH/$1-$BUILD_NUMBER-$2.txt
	
	echo_value "Xcode Output" $XCODE_OUTPUT
	xcodebuild \
			-configuration $XCODE_CONFIGURATION \
			-target $XCODE_TARGET \
			build \
			OBJROOT=$OBJ_ROOT \
			SYMROOT=$SYM_ROOT \
			BUILD_NUMBER=$BUILD_NUMBER \
			> "$XCODE_OUTPUT"
}


# Copies the mobileprovision file out of the app package
copy_mobileprovision() {
	local XCODE_TARGET=$1
	local XCODE_CONFIGURATION=$2
	local XCODE_ARCH=iphoneos

	local PRODUCTS_PATH="$SYM_ROOT/$XCODE_CONFIGURATION-$XCODE_ARCH"
	local PROVISION_PATH="$PRODUCTS_PATH/$XCODE_TARGET.mobileprovision"
	
	cp $PRODUCTS_PATH/*.app/embedded.mobileprovision $PROVISION_PATH
}


# Creates a zip of the contents of the build products directory
make_zip() {
	local XCODE_TARGET=$1
	local XCODE_CONFIGURATION=$2
	local XCODE_ARCH=iphoneos

	local PRODUCTS_PATH="$SYM_ROOT/$XCODE_CONFIGURATION-$XCODE_ARCH"
	local BUILD_NAME="$XCODE_TARGET-$BUILD_NUMBER-$XCODE_CONFIGURATION"
	local ZIP_PATH="$OUTPUT_PATH/$BUILD_NAME.zip"
	local DSYM_PATH="$OUTPUT_PATH/$BUILD_NAME.dSYM"

	# save the dSYM file, then move it out of the way
	mkdir $DSYM_PATH
	cp -r $PRODUCTS_PATH/*.dSYM $DSYM_PATH
	rm -r $PRODUCTS_PATH/*.dSYM
	echo_value "Saved dSYM" $DSYM_PATH

	# use `ditto` because iTunes Connect will not accept output of `zip`
	ditto -c -k -v --sequesterRsrc $PRODUCTS_PATH $ZIP_PATH
	echo_value "Created Archive" $ZIP_PATH
}


# Clean the build directories
clean() {
	rm -r $OBJ_ROOT/* $SYM_ROOT/*
}


# GENERAL SUPPORT FUNCTIONS


# Fancy echo of a label and a value
function echo_value() {
	local ANSI_BG_GRAY='\E[47m'
	local ANSI_OFF='\E[0m'
	echo -e $1: $ANSI_BG_GRAY$2$ANSI_OFF
}


# Creates a RAM disk device, formats, then mounts it.
function create_ram_disk() {
	local RAMDISK_SECTORS=$((2048 * $RAMDISK_SIZE_MB))
	RAMDISK_DEVICE=`hdiutil attach -nomount ram://$RAMDISK_SECTORS`
	RAMDISK_PATH=`mktemp -d /tmp/ramdisk.XXXXXX`
	newfs_hfs $RAMDISK_DEVICE # format as HFS+
	mount -t hfs $RAMDISK_DEVICE $RAMDISK_PATH
	df -h $RAMDISK_PATH # report on disk usage
}


# Destroys the RAM disk created by create_ram_disk
function destroy_ram_disk() {
	echo_value "Destroying RAM Disk" $RAMDISK_DEVICE
	df -h $RAMDISK_PATH # report on disk usage
	umount -f $RAMDISK_DEVICE
	hdiutil detach $RAMDISK_DEVICE
	rmdir $RAMDISK_PATH
	unset RAMDISK_DEVICE
	unset RAMDISK_PATH
}


function generate_build() {
	OUTPUT_PATH=$OUTPUT_ROOT/$TAG_PREFIX$BUILD_NUMBER
	if [ -e "$OUTPUT_PATH" ]; then
		echo_value Deleting $OUTPUT_PATH
		rm -rf $OUTPUT_PATH
	fi
	echo_value Creating $OUTPUT_PATH
	mkdir -p $OUTPUT_PATH
	
	create_ram_disk
	SRC_ROOT=$RAMDISK_PATH/Source
	OBJ_ROOT=$RAMDISK_PATH/Intermediate
	SYM_ROOT=$RAMDISK_PATH/Products

	FILESLIST_PATH=$OUTPUT_PATH/Files-$BUILD_NUMBER.txt

	case $SCM in
		git)
			mkdir $SRC_ROOT
			cd ..
			git archive --format=tar $GIT_BUILD_TAG^{tree} | \
				tar xvf - -C $SRC_ROOT > $FILESLIST_PATH
			;;
		svn)
			svn export $SVN_URL_BUILD $SRC_ROOT > $FILESLIST_PATH
			;;
	esac

	cd $SRC_ROOT
	do_build # defined in build-config.sh
	cd $OUTPUT_PATH
	destroy_ram_disk
}


# GIT-SPECIFIC FUNCTIONS


# Determines the tag of the next build
function determine_git_build_tag() {
	local LAST_TAG=`git tag -l "${TAG_PREFIX}????" | sort | tail -n 1`

	if [ -z "$LAST_TAG" ]; then
		echo_value "Last Tag" "not defined"
		LAST_TAG=${TAG_PREFIX}0000
	else
		echo_value "Last Tag" $LAST_TAG
	fi
	
	local LAST_NUMBER=${LAST_TAG#$TAG_PREFIX}
	echo_value "Last Number" $LAST_NUMBER
	
	BUILD_NUMBER=`printf "%04d" $((${LAST_NUMBER##*0}+1))`
	echo_value "This Number" $BUILD_NUMBER
	
	GIT_BUILD_TAG=$TAG_PREFIX$BUILD_NUMBER
	
	git tag -a -m "tagging build $BUILD_NUMBER" $GIT_BUILD_TAG
}


function git_log_by_build() {
	if [ "$1" -lt 1 ]; then
		echo Impossible build number: $1
		exit 1
	fi
	if [ "$2" -lt 1 ]; then
		echo Impossible build number: $2
		exit 1
	fi

	BUILD_NUMBER_A=`printf "%04d" $1`
	exit_on_nonzero_status
	BUILD_NUMBER_B=`printf "%04d" $2`
	exit_on_nonzero_status
	
	TAG_A=$TAG_PREFIX$BUILD_NUMBER_A
	TAG_B=$TAG_PREFIX$BUILD_NUMBER_B

	git log $TAG_A..$TAG_B
}


# SVN-SPECIFIC FUNCTIONS


# Determines the Root URL of the Subversion Repository
function determine_svn_url_root() {
	local SVN_ROOT_LINE=`svn info | grep "^Repository Root: "`
	if [ -z "$SVN_ROOT_LINE" ]; then
		exit
	fi
	
	SVN_URL_ROOT=${SVN_ROOT_LINE#Repository Root: }
	SVN_URL_TRUNK=$SVN_URL_ROOT/$SVN_PATH_TRUNK
	SVN_URL_TAGS=$SVN_URL_ROOT/$SVN_PATH_TAGS

	echo_value "Repository Root" $SVN_URL_ROOT
}


# Determines the URL of the next build
function determine_svn_url_build() {
	local LAST_TAG=`svn list $SVN_URL_TAGS | grep "^$TAG_PREFIX" | sort | tail -n 1`
	
	if [ -z "$LAST_TAG" ]; then
		echo_value "Last Tag" "not defined"
		LAST_TAG=${TAG_PREFIX}0000/
	else
		echo_value "Last Tag" $LAST_TAG
	fi
	
	local LAST_TAG_SUFFIX=${LAST_TAG#$TAG_PREFIX}
	local LAST_NUMBER=${LAST_TAG_SUFFIX%/}
	echo_value "Last Number" $LAST_NUMBER
	
	# strip leading zeroes
	X=${LAST_NUMBER}
	Y=${LAST_NUMBER#0}
	while [ $X != $Y ]; do
		X=$Y
		Y=${Y#0}
	done
	
	BUILD_NUMBER=`printf "%04d" $(($X+1))`
	echo_value "This Number" $BUILD_NUMBER
	
	SVN_URL_BUILD=$SVN_URL_TAGS/$TAG_PREFIX$BUILD_NUMBER

	svn copy --quiet --message "tagging build $BUILD_NUMBER" --parents \
		$SVN_URL_TRUNK $SVN_URL_BUILD
}


function svn_log_by_build() {
	if [ "$1" -lt 1 ]; then
		echo Impossible build number: $1
		exit 1
	fi
	if [ "$2" -lt 1 ]; then
		echo Impossible build number: $2
		exit 1
	fi

	BUILD_NUMBER_A=`printf "%04d" $1`
	exit_on_nonzero_status
	BUILD_NUMBER_B=`printf "%04d" $2`
	exit_on_nonzero_status

	determine_svn_url_root

	SVN_URL_BUILD_A=$SVN_URL_TAGS/$TAG_PREFIX$BUILD_NUMBER_A
	LINE=`svn info $SVN_URL_BUILD_A | grep "^Last Changed Rev:"`
	SVN_REV_A=${LINE#Last Changed Rev: }
	
	SVN_URL_BUILD_B=$SVN_URL_TAGS/$TAG_PREFIX$BUILD_NUMBER_B
	LINE=`svn info $SVN_URL_BUILD_B | grep "^Last Changed Rev:"`
	SVN_REV_B=${LINE#Last Changed Rev: }

	echo_value "Build A" $BUILD_NUMBER_A
	echo_value "Build B" $BUILD_NUMBER_B

	echo_value "Revision A" $SVN_REV_A
	echo_value "Revision B" $SVN_REV_B
	
	svn log --verbose --revision $SVN_REV_A:$SVN_REV_B $SVN_URL_ROOT
}


# MAIN PROGRAM FLOW


function echo_usage_and_exit() {
	cat <<USAGE_TEXT;
Usage:
    mkbuild new
        create a new build
    mkbuild rebuild N
        recreate build N
    mkbuild log N
        print changes between build N-1 and build N
    mkbuild log N M
        print changes between build N and build M
    mkbuild list
        list all builds in repository
    mkbuild archive N
    	create a zip archive of the source of build N
USAGE_TEXT
	exit 2
}


function exit_on_nonzero_status() {
	if [ $? -ne 0 ]; then
		echo_usage_and_exit
	fi
}


function do_list() {
	case $SCM in
		git)
			git tag -l "${TAG_PREFIX}????"
			;;
		svn)
			determine_svn_url_root
			svn list $SVN_URL_TAGS
			;;
	esac
}


function do_log() {
	case $SCM in
		git)
			if [ -z "$2" ]; then
				git_log_by_build $(($1 - 1)) $1
			else
				git_log_by_build $1 $2
			fi
			;;
		svn)
			;;
	esac
}


function do_new() {
	case $SCM in
		git)
			determine_git_build_tag
			;;
		svn)
			determine_svn_url_root
			determine_svn_url_build
			;;
	esac
	generate_build
}


function do_rebuild() {
	BUILD_NUMBER=`printf "%04d" $1`
	exit_on_nonzero_status
	echo_value "Rebuilding Number" $BUILD_NUMBER
	
	case $SCM in
		git)
			GIT_BUILD_TAG=$TAG_PREFIX$BUILD_NUMBER
			echo "NOT IMPLEMENTED!!!"
			;;
		svn)
			determine_svn_url_root
			SVN_URL_BUILD=$SVN_URL_TAGS/$TAG_PREFIX$BUILD_NUMBER
			;;
	esac
}


function do_archive() {
	BUILD_NUMBER=`printf "%04d" $1`
	exit_on_nonzero_status
	echo_value "Archiving Number" $BUILD_NUMBER

	ZIP_PATH=Source-$BUILD_NUMBER.zip

	case $SCM in
		git)
			GIT_BUILD_TAG=$TAG_PREFIX$BUILD_NUMBER
			git archive --format=zip $GIT_BUILD_TAG^{tree} > $ZIP_PATH
			;;
		svn)
			determine_svn_url_root
			SVN_URL_BUILD=$SVN_URL_TAGS/$TAG_PREFIX$BUILD_NUMBER
			create_ram_disk
			SRC_ROOT=$RAMDISK_PATH/Source
			svn export --quiet $SVN_URL_BUILD $SRC_ROOT
			ditto -c -k -v --sequesterRsrc $SRC_ROOT $ZIP_PATH
			destroy_ram_disk
			;;
	esac
	
	echo_value "Created Archive" $ZIP_PATH
}


# MAIN


if [ ! -e build-config.sh ]; then
	echo To use mkbuild, you must create a build-config.sh file in the current
	echo directory.
	exit 3
fi

source build-config.sh

case "$SCM" in
	svn | git)
		;;
	none)
		echo Source control system not defined.  Please include SCM=svn or
		echo SCM=git in your build-config.sh file.
		exit 3
		;;
	*)
		echo Unknown source control system: $SCM
		exit 3
		;;
esac

case "$1" in
	list | ls)
		do_list
		;;
	log)
		do_log $2 $3
		;;
	new)
		do_new
		;;
	archive)
		do_archive $2
		;;
	rebuild)
		do_rebuild $2
		;;
	*)
		echo_usage_and_exit
		;;
esac

# THE END
